!=======================================================================
!
! MODULE  : odedef.F
!
! PURPOSE : Problem-specific routines for frame.F
!
! CONTAINS:
!           subroutine setproblem
!           subroutine initproblem
!           subroutine derivs
!           subroutine jacobian
!
!=======================================================================


!=======================================================================
!
! SUBROUTINE: setproblem
!
! PURPOSE   : Determines basic parameters for integrators depending on problem.
!
! SOURCE    :
!
! INPUT     : iprob : problem number
!
! OUTPUT    : nvar  : number of ODEs
!             narg  : number of additional arguments for derivs function
!
!=======================================================================

subroutine setproblem(iprob, nvar, narg, logunit)

  implicit none

! subroutine arguments
  integer :: nvar, narg, iprob, logunit

!-----------------------------------------------------------------

  select case (iprob)
    case(1) ! (d/dx)y = y
      nvar = 1
      narg = 1+1
    case(2) ! stiff problem (1)
      write(logunit,'("SETPROBLEM: problem 2: nothing there yet")')
    case(3) ! stiff problem (2)
      nvar = 3
      narg = 3+1
    case default
      write(logunit,'("SETPROBLEM: invalid iprob")')
      stop
  end select

  return
end

!=======================================================================


!=======================================================================
!
! SUBROUTINE: initproblem
!
! PURPOSE   : Determines basic parameters for integrators depending on problem.
!
! SOURCE    :
!
! INPUT     : iprob : problem number
!             nvar  : number of ODEs
!             narg  : number of additional arguments for derivs function
!
! OUTPUT    : ystart: initial conditions
!             argarr: arguments for derivs (constant coefficients)
!
!=======================================================================

subroutine initproblem(iprob, nvar, narg, ystart, argarr, inpunit, logunit, xstart)

  implicit none

! subroutine arguments
  integer          :: nvar, narg, iprob, inpunit, logunit
  double precision :: ystart(nvar)
  double precision :: argarr(narg)

! local variables
  double precision :: xstart

!-----------------------------------------------------------------

  argarr(1) = dble(iprob) ! This we need for derivs to chose the correct right-hand side.

  select case (iprob)
    case(1) ! (d/dx)y = a*y, x=[xstart,10]
      argarr(2) = 2.5d0
      ystart(1) = 1.d-3*exp(2.5*xstart)
    case(2) ! stiff problem, x=[0,4]
      write(logunit,'("SETPROBLEM: problem 2: nothing there yet")')
    case(3) ! stiff problem (2), x=[0,50]
      argarr(2) = -1.3d-2
      argarr(3) = -1.0d3
      argarr(4) = -2.5d3
      ystart(1) =  1.d0
      ystart(2) =  1.d0
      ystart(3) =  0.d0
    case default
      write(logunit,'("INITPROBLEM: invalid iprob")')
      stop
  end select

  return
end

!=======================================================================


!=======================================================================
!
! SUBROUTINE: derivs
!
! PURPOSE   : Computes RHS of nvar ODEs, i.e., the derivatives.
!
! SOURCE    :
!
! INPUT     : nvar   : number of ODEs
!             x      : independent variable
!             y      : dependent variables
!             narg   : number of additional arguments
!             argarr : additional arguments (constant factors in ODEs)
!
! OUTPUT    : dydx   : derivatives (RHS)
!
!=======================================================================

subroutine derivs(nvar, x, y, narg, argarr, dydx)

  implicit none

! subroutine arguments
  integer          :: nvar, narg
  double precision :: x
  double precision :: y     (nvar)
  double precision :: dydx  (nvar)
  double precision :: argarr(narg)

! local variables
  integer          :: iprob, logunit

!-----------------------------------------------------------------

  iprob = int(argarr(1)) ! This we need for derivs to chose the correct right-hand side.

  select case (iprob)
    case(1) ! (d/dx)y = a*y
      dydx(1) = argarr(2)*y(1)
    case(2) ! stiff problem (1)
      write(logunit,'("SETPROBLEM: problem 2: nothing there yet")')
    case(3) ! stiff problem (2)
      dydx(1) = argarr(2)*y(1) + argarr(3)*y(1)*y(3)
      dydx(2) = argarr(4)*y(2)*y(3)
      dydx(3) = argarr(2)*y(1) + argarr(3)*y(1)*y(3) + argarr(4)*y(2)*y(3)
    case default ! invalid iprob
      write(logunit,'("DERIVS: invalid iprob")')
      stop
  end select

  return
end

!=======================================================================


!=======================================================================
!
! SUBROUTINE: jacobian
!
! PURPOSE   : Computes df/dx and df/dy, where f is the RHS of the ODEs.
!             df/dx = 0 if there is no explicit dependence of the RHS on x.
!             In dfdy, rows (first index) give f(i), and
!                      columns (second index) denote derivative w.r.t. y(i)
!
! SOURCE    :
!
! INPUT     : nvar   : number of ODEs
!             x      : independent variable
!             y      : dependent variables
!             narg   : number of additional arguments
!             argarr : additional arguments (constant factors in ODEs)
!
! OUTPUT    : dfdy   : Jacobian
!             dfdx   : derivative of RHS w.r.t x
!
!=======================================================================

subroutine jacobian(nvar, x, y, narg, argarr, dfdx, dfdy)

  implicit none

! subroutine arguments
  integer          :: nvar, narg
  double precision :: x
  double precision :: y     (nvar)
  double precision :: dfdx  (nvar)
  double precision :: dfdy  (nvar,nvar)
  double precision :: argarr(narg)

! local variables
  integer          :: iprob, logunit

!-----------------------------------------------------------------

  iprob = int(argarr(1)) ! This we need for derivs to chose the correct right-hand side.

  select case (iprob)
    case(1) ! (d/dx)y = a*y
      dfdx(1)   = 0.d0
      dfdy(1,1) = argarr(2)
    case(2) ! stiff problem (1)
      write(logunit,'("SETPROBLEM: problem 2: nothing there yet")')
    case(3) ! stiff problem (2)
      dfdx(:)   = 0.d0
      dfdy(1,1) = argarr(2) + argarr(3)*y(3)
      dfdy(1,2) = 0.d0
      dfdy(1,3) = argarr(3)*y(1)
      dfdy(2,1) = 0.d0
      dfdy(2,2) = argarr(4)*y(3)
      dfdy(2,3) = argarr(4)*y(2)
      dfdy(3,1) = argarr(2) + argarr(3)*y(3)
      dfdy(3,2) = argarr(4)*y(3)
      dfdy(3,3) = argarr(3)*y(1) + argarr(4)*y(2)
    case default ! invalid iprob
      stop
  end select

  return
end

!=======================================================================

