!=======================================================================
!
! SUBROUTINE: odeint.F
!
! PURPOSE   : Driver for ODE integrators
!
! SOURCE    : Numerical Recipes, Press et al. 2nd ed. 1992
!
! INPUT     : nvar   : number of ODEs
!             maxstep: maximum number of steps. If reached, routine will stop.
!             kmax   : If > 0, intermediate results will be stored in xp, yp (see below).
!                      If = 0, only end values will be returned (in ystart)
!             kcount : number of actual steps taken. Determines "active" length of xp, yp.
!             iwhich : switch for integrator/stepper
!                        0: Euler: h1 must be (x2-x1)/maxstep
!                        1: RK4, fixed stepsize. Same restrictions as Euler.
!                        2: RK5, adaptive stepsize (rkqs).
!                        3: Kapps-Rosenbrock semi-implicit
!             x1, x2 : start and end point of integration
!             eps    : desired accuracy
!             h1     : initial guess for step size
!             hmin   : minimum allowed step size (can be zero)
!             dxsav  : storage intervals. If 0.0, all steps are used.
!                      dxsav should be positive.
!             nok    : number of good steps
!             nbad   : number of bad (but retried and fixed) steps
!             ystart : start values of y. Will be replaced by end values upon return.
!             narg   : number of problem-specific arguments for derivs
!             argarr : problem-specific coefficients for derivs
!             xp     : if kmax > 0, contains intermediate x [real xp(1:kmax)]
!             yp     : if kmax > 0, contains intermediate y [real yp(1:kmax,1:nvar)]
!                      xp and yp need to be defined in calling routine.
!             done   : flag indicating exit status:
!                        1: normal exit (integration ok)
!                        2: step size too small
!                        3: step overflow (nstp > maxstep)
!             logunit
!
! OUTPUT    : xp     : intermediate values x
!             yp     : intermediate values y
!             ystart : end values
!
! EXTERNAL  : subroutine derivs(nvar, x, y, narg, argarr, dydx)
!             subroutine eulerstep
!             subroutine rk4
!             subroutine rkqs(nvar   ,                  &
!                             x      , htry   , eps   , &
!                             y      , dydx   , yscal , &
!                             narg   , argarr ,         &
!                             hdid   , hnext  ,         &
!                             logunit)
!
! USAGE     : Given y at x1 (ystart), integrate the ODEs defined by
!             derivs using adaptive step size. The stepping routine's
!             name is rkqs, but can contain whatever stepper seems
!             adequate.
!             Note that eulerstep and rk4 get an additional parameter
!             hf ("h-fixed")
!
!=======================================================================

subroutine odeint (nvar   , maxstep, kmax  , kcount, iwhich ,          &
                   x1     , x2     , eps   , h1    , hmin   , dxsav  , &
                   nok    , nbad   ,                                   &
                   ystart ,                                            &
                   narg   , argarr ,                                   &
                   xp     , yp     ,                                   &
                   done   , logunit)

  implicit none
  integer, parameter :: i8 = selected_int_kind(18)

! subroutine arguments
  integer(kind=i8) :: maxstep
  integer          :: nvar, narg, kmax, kcount, logunit, nok, nbad, done, iwhich
  double precision :: x1, x2, eps, h1, hf, hmin, dxsav
  double precision :: ystart(nvar)
  double precision :: xp(kmax)      ! Unfortunately, fortran does not allow 0 here;
  double precision :: yp(kmax,nvar) ! so, if kmax=0, we still have to define these arrays.
  double precision :: argarr(narg)

! local variables
  integer(kind=i8)              :: nstp
  double precision              :: x, xsav, hnext, hdid, htry, h, hsav, hstep
  double precision, allocatable :: yscal(:) ! scaling vector for errors
  double precision, allocatable :: y    (:) ! temporary result for given step
  double precision, allocatable :: dydx (:) ! derivatives
  double precision, allocatable :: yout (:) ! auxiliary

  double precision, parameter   :: tiny = 1.0d-30

!-----------------------------------------------------------------

! allocate local vectors
  allocate(yscal(nvar))
  allocate(y    (nvar))
  allocate(dydx (nvar))
  allocate(yout (nvar))

  x      = x1
  nok    = 0
  nbad   = 0
  kcount = 0
  y(:)   = ystart(:)

  if (kmax .gt. 0) xsav = x - dxsav*2.d0   ! assures storage of first step
!
!-----------------------------------------------------------------
! at first, the fixed-step integrators

  if (iwhich .lt. 2) then
    dxsav  = dxsav*(1.d0-1.d-13) ! to avoid truncation errors
    hf = (x2-x1)/dble(maxstep) ! for fixed-step integrators!
    h  = hf
    write(logunit,'("ODEINT: fixed timestep h = ",e23.15)') hf

    do nstp = 1,maxstep

      call derivs(nvar, x, y, narg, argarr, dydx)
      if (kmax .gt. 0) then
        if (abs(x-xsav) .ge. abs(dxsav)) then
! store intermediate results
          if (kcount .lt. kmax-1) then
            kcount = kcount + 1
            xp(kcount) = x
            yp(kcount,:) = y(:)
            xsav = x
          end if
        end if
      end if

      if(iwhich .eq. 0) then
        call eulerstep(nvar, x, h, y, dydx, yout, logunit)
      else
        call rk4(nvar, x, h, y, dydx, narg, argarr, yout, logunit)
      endif
      y = yout

      write(logunit,'("ODEINT: nstp=",i8," h=",e23.15," x=",e23.15," y(1)=",e23.15)') &
        nstp, h, x, y(1)

      nok = nok + 1
    enddo

    if (kmax .ne. 0) then
! save final step
      kcount = kcount + 1
      xp(kcount) = x
      yp(kcount,:) = y(:)
    end if

    done = 1 ! success
    return
  endif

  if (iwhich .lt. 2) stop ' wrong path'
!
!--------------------------------------------------------------------------
! now, integration with step-size control
!

  h     = dsign(h1,x2-x1)    ! chose correct step direction
  dxsav = dsign(dxsav,x2-x1) ! chose correct step direction


itloop: do nstp = 1,maxstep

    call derivs(nvar, x, y, narg, argarr, dydx)

! scaling for monitoring accuracy. Change if necessary.
    yscal(:) = abs(y(:)) + abs(h*dydx(:)) + tiny
!   yscal(:) = dmax1(1.d0,dabs(y(:))) ! according to Fabian

    if (kmax .gt. 0) then
      if (abs(x-xsav) .ge. abs(dxsav)) then
! store intermediate results
        if (kcount .lt. kmax-1) then
          kcount = kcount + 1
          xp(kcount) = x
          yp(kcount,:) = y(:)
          xsav = x
        end if
      end if
    end if

! to obtain intermediate grids with a TYPICAL resolution of dxsav,
! uncomment following line
!   if ((h-dxsav)* (x+h-x1).gt.0.d0) h = dxsav*(1.d0+1.d-13)

! if stepsize can overshoot, decrease
    if ((x+h-x2)*(x+h-x1) .gt. 0.d0) h = x2 - x

    if (iwhich .eq. 2) then
      call rkqs(nvar, x, h, eps, y, dydx, yscal, narg, argarr, hdid , hnext, logunit)
    else if (iwhich .eq. 3) then
      call stiff(nvar, x, h, eps, y, dydx, yscal, narg, argarr, hdid, hnext, logunit)
    else
      write(logunit,'("ODEINT: invalid iwhich")')
      stop
    endif
    write(logunit,'("ODEINT: nstp=",i8," h=",e23.15," x=",e23.15," y(1)=",e23.15)') &
      nstp, h, x, y(1)

    if (hdid .eq. h) then
      nok = nok + 1
    else
      nbad = nbad + 1
    end if

    if ((x-x2)*(x2-x1) .ge. 0.d0) then
      ystart = y
      if (kmax.ne.0) then
! save final step
        kcount = kcount + 1
        xp(kcount) = x
        yp(kcount,:) = y(:)
      end if
      done = 1 ! success
      return
    end if

    if (abs(hnext) .lt. hmin) then
      write(logunit,'("ODEINT: step size too small: x,hnext=",e23.15,e23.15)') &
        x, hnext
      done = 2
      return
    endif

    h = hnext

  end do itloop

  write(logunit,'("ODEINT: Step overflow at x = ",e23.15)') x
  done = 3

return
end

!=======================================================================

