!=======================================================================
!
! SUBROUTINE: rkck.F
!
! PURPOSE   : Performs one fifth-order Cash-Karp RK-step.
!
! SOURCE    : Numerical Recipes, Press et al. 2nd ed. 1992
!
! INPUT     : nvar   : number of ODEs
!             x      : current x (independent variable)
!             h      : step size over which to advance the solution
!             y      : current values of ODEs
!             dydx   : derivatives at x (and work array)
!             narg   : number of problem-specific arguments for derivs
!             argarr : problem-specific coefficients for derivs
!             logunit
!
! OUTPUT    : yout   : incremented values of ODEs (new solution)
!             yerr   : local truncation error
!
! EXTERNAL  : subroutine derivs(nvar, x, y, narg, argarr, dxdy)
!
! USAGE     : Given y at x, advances solution by step size h using the
!             derivatives in the user-supplied subroutine derivs.
!
!=======================================================================

subroutine rkck(nvar   ,                &
                x      , h      ,       &
                y      , dydx   ,       &
                narg   , argarr ,       &
                yout   ,                &
                yerr   ,                &
                logunit)

  implicit none

! subroutine arguments
  integer          :: nvar, narg, logunit
  double precision :: x, h
  double precision :: y     (nvar)
  double precision :: dydx  (nvar)
  double precision :: argarr(narg)
  double precision :: yout  (nvar)
  double precision :: yerr  (nvar)

! local variables
  integer                       :: i
  double precision, allocatable :: ak2(:), ak3(:), ak4(:), ak5(:), ak6(:)
  double precision, allocatable :: ytemp(:)

  double precision, parameter ::   &
    a2  =          0.2d0,          & !  1/5
    a3  =          0.3d0,          & !  3/10
    a4  =          0.6d0,          & !  3/5
    a5  =          1.0d0,          & !  1
    a6  =          0.875d0,        & !  7/8
    b21 =          0.2d0,          & !  1/5
    b31 =          0.075d0,        & !  3/40
    b32 =          0.225d0,        & !  9/40
    b41 =          0.3d0,          & !  3/10
    b42 =         -0.9d0,          & ! -9/10
    b43 =          1.2d0,          & !  6/5
    b51 =        -11.d0/54.d0,     &
    b52 =          2.5d0,          & !  5/2
    b53 =        -70.d0/27.d0,     &
    b54 =         35.d0/27.d0,     &
    b61 =       1631.d0/55296.d0,  &
    b62 =        175.d0/512.d0,    &
    b63 =        575.d0/13824.d0,  &
    b64 =      44275.d0/110592.d0, &
    b65 =        253.d0/4096.d0,   &
    c1  =         37.d0/378.d0,    &
    c3  =        250.d0/621.d0,    &
    c4  =        125.d0/594.d0,    &
    c6  =        512.d0/1771.d0,   &
    dc1 = c1 -  2825.d0/27648.d0,  &
    dc3 = c3 - 18575.d0/48384.d0,  &
    dc4 = c4 - 13525.d0/55296.d0,  &
    dc5 =    -   277.d0/14336.d0,  &
    dc6 = c6 -     0.25d0            !  1/4

!-----------------------------------------------------------------

! allocate local vectors
  allocate(ak2  (nvar))
  allocate(ak3  (nvar))
  allocate(ak4  (nvar))
  allocate(ak5  (nvar))
  allocate(ak6  (nvar))
  allocate(ytemp(nvar))

  ytemp(:) = y(:) + dydx(:)*b21*h                                          ! first step
  call derivs(nvar, x+a2*h, ytemp, narg, argarr, ak2)                      ! second step
  ytemp(:) = y(:) + (dydx(:)*b31 + ak2(:)*b32)*h
  call derivs(nvar, x+a3*h, ytemp, narg, argarr, ak3)                      ! third step
  ytemp(:) = y(:) + (dydx(:)*b41 + ak2(:)*b42 + ak3(:)*b43)*h
  call derivs(nvar, x+a4*h, ytemp, narg, argarr, ak4)                      ! fourth step
  ytemp(:) = y(:) + (dydx(:)*b51 + ak2(:)*b52 + ak3(:)*b53 + ak4(:)*b54)*h
  call derivs(nvar, x+a5*h, ytemp, narg, argarr, ak5)                      ! fifth step
  ytemp(:) = y(:) + (dydx(:)*b61 + ak2(:)*b62 + ak3(:)*b63 + ak4(:)*b64 + ak5(:)*b65)*h
  call derivs(nvar, x+a6*h, ytemp, narg, argarr, ak6)                      ! sixth step
  yout (:) = y(:) + (dydx(:)*c1  + ak3(:)*c3  + ak4(:)*c4               + ak6(:)*c6 )*h
  yerr (:) =        (dydx(:)*dc1 + ak3(:)*dc3 + ak4(:)*dc4 + ak5(:)*dc5 + ak6(:)*dc6)*h

  deallocate(ytemp)
  deallocate(ak6)
  deallocate(ak5)
  deallocate(ak4)
  deallocate(ak3)
  deallocate(ak2)

  return
end

!=======================================================================

